/* This file is part of the "Regression Tree Fields" (RTF) source code distribution,
 * obtained from http://research.microsoft.com/downloads.
 * It is provided to you under the terms of the Microsoft Research License Agreement
 * (MSR-LA). Please see License.txt for details.
 *
 *
 * File: Dataset.h
 * Implements the dataset class of the multi-label clasification example.
 *
 */

#ifndef H_MULTILABEL_DATASET_H
#define H_MULTILABEL_DATASET_H

#include <map>

#include <boost/filesystem.hpp>

#include "RTF/Image.h"
#include "RTF/Array.h"
#include "RTF/Types.h"
#include "RTF/Utility.h"
#include "RTF/Training.h"

namespace MultiLabel
{

    class IOException : public std::exception
    {
    private:
        std::string w;
    public:
        IOException(const std::string& what_) : w(what_) {}

        virtual const char *what() const throw()
        {
            return w.c_str();
        }

        virtual ~IOException() throw() {}
    };

    class Scene;
    class Yeast;
    class Synth;

    template<typename TypeTag = Scene>
    struct DatasetTraits
    {
        static const size_t NumClasses  = 6;
        static const size_t NumFeatures = 294;
    };

    template<>
    struct DatasetTraits<Yeast>
    {
        static const size_t NumClasses  = 14;
        static const size_t NumFeatures = 103;
    };

    template<>
    struct DatasetTraits<Synth>
    {
        static const size_t NumClasses  = 3;
        static const size_t NumFeatures = 3;
    };


    // The dataset class for the multi-label task; methods and typedefs marked by INTERFACE
    // are required by the RTF code and must be implemented by any dataset class.
    template <typename TypeTag = Scene>
    class Dataset
    {
    public:
        typedef DatasetTraits<TypeTag> Traits;

        // INTERFACE: define the output label of each variable to be binary
        typedef Training::LabelVector<double, 2> UnaryGroundLabel;
        typedef Training::LabelVector<double, 4> PairwiseGroundLabel;

        // INTERFACE: define the type of the input data stored for each variable
        typedef Training::LabelVector<double, Traits::NumFeatures>  InputLabel;

        Dataset(const std::string& path_, const std::string& type_,
                size_t ourPart_ = 0, size_t numParts_ = 1, bool useOtherParts_ = false)
            : path(path_), type(type_), ourPart(ourPart_), numParts(numParts_), useOtherParts(useOtherParts_)
        {
            ReadDescriptor();
        }

        std::string Type() const
        {
            return type;
        }

        // INTERFACE: return the number of images in the dataset
        size_t GetImageCount() const
        {
            return allLabels.size();
        }

        // INTERFACE: return the idx'th ground truth image
        ImageRefC<UnaryGroundLabel> GetGroundTruthImage(size_t idx) const
        {
            ImageRef<UnaryGroundLabel> ret(Traits::NumClasses, 1);
            auto &labels = allLabels[idx];

            // First index of ground truth label set to 1 indicates label is not present
            for( size_t v = 0; v < Traits::NumClasses; ++v )
            {
                ret(v, 0)[0] = 1.0;
                ret(v, 0)[1] = 0.0;
            }

            // Second index of ground truth label set to 1 indicates label is present
            for( size_t i = 0; i < labels.size(); ++i )
            {
                const auto v = labels[i];
                ret(v, 0)[0] = 0.0;
                ret(v, 0)[1] = 1.0;
            }

            return ret;
        }

        // INTERFACE: return the idx'th input image.
        ImageRefC<InputLabel> GetInputImage(size_t idx) const
        {
            ImageRef<InputLabel> img(Traits::NumClasses, 1);
            auto &features = allFeatures[idx];

            for( size_t v = 0; v < Traits::NumClasses; ++v )
            {
                std::copy(features.begin(), features.end(), &(img(v, 0)[0]));
            }

            return img;
        }

    private:

        void ReadDescriptor()
        {
            boost::filesystem::path dpath = path;
            dpath /= (type + ".svm");
            std::ifstream ifs(dpath.string());

            if( ifs.fail() )
                throw IOException("failed to open '" + dpath.string() + "'");

            std::string line;
            int lineidx = -1;
            while(std::getline(ifs, line))
            {
                lineidx++;

                if( useOtherParts  &&  lineidx % numParts == ourPart )
                    continue;

                if( !useOtherParts &&  lineidx % numParts != ourPart )
                    continue;

                std::stringstream linestream(line);
                std::string token;
                std::vector<size_t> labels;
                std::vector<double> features;
                while( std::getline(linestream, token, ' ') )
                {
                    if( labels.size() == 0 )
                    {
                        std::stringstream labelstream(token);
                        std::string label;
                        while( std::getline(labelstream, label, ',') )
                        {
                            auto labelid = std::atoi(label.c_str());
                            if( labelid >= Traits::NumClasses )
                            {
                                std::cerr << "Unexpected label id: " << labelid << std::endl;
                                throw IOException("Unexpected label id");
                            }
                            labels.push_back(labelid);
                        }
                    }
                    else
                    {
                        // assume features are ordered and dense, starting at the first index
                        size_t validx = token.find(':') + 1;
                        std::string valstr(&token[validx], token.length()-validx);
                        features.push_back(std::atof(valstr.c_str()));
                    }
                }
                if( features.size() != Traits::NumFeatures )
                {
                    std::cerr << "Unexpected number of features" << std::endl;
                    throw IOException("Unexpected number of features");
                }
                allFeatures.push_back(features);
                allLabels.push_back(labels);
            }
        }

        const std::string path;
        const std::string type;

        size_t ourPart;
        size_t numParts;
        bool useOtherParts;

        std::vector< std::vector<double> > allFeatures;
        std::vector < std::vector<size_t> > allLabels;
    };
}

#endif // H_MULTILABEL_DATASET_H
