/* This file is part of the "Regression Tree Fields" (RTF) source code distribution,
 * obtained from http://research.microsoft.com/downloads.
 * It is provided to you under the terms of the Microsoft Research License Agreement
 * (MSR-LA). Please see License.txt for details.
 *
 *
 * File: Features.h
 * Implements the feature class of the multi-label classification example.
 *
 */

#ifndef H_MULTILABEL_FEATURES_H
#define H_MULTILABEL_FEATURES_H

#include <random>
#include <algorithm>

#include <Eigen/Eigen>

#include "RTF/Compute.h"

#include "Dataset.h"

namespace MultiLabel
{
    // The feature class in this example is fairly minimal, since we do not actually train
    // trees. Hence, only the linear basis functionality is really used.
    template <typename TypeTag = Scene>
    class Feature
    {
    public:
        typedef Dataset<TypeTag> DS;
        typedef typename Dataset<TypeTag>::InputLabel InputLabel;

        // INTERFACE: the type of a pre-processed instance (can be the same as the input)
        typedef ImageRefC<InputLabel> PreProcessType;

        // INTERFACE: linear basis function support - dimensionality of the bases
        static const size_t UnaryBasisSize = InputLabel::Size;
        static const size_t PairwiseBasisSize = 1;

        // INTERFACE: return a prep-processed input instance of type PreProcessType
        static PreProcessType PreProcess(const ImageRefC<InputLabel>& input)
        {
            return input;
        }

        // INTERFACE: fill in the linear basis
        static void ComputeBasis(int x, int y, const PreProcessType& prep,
                                 const VecCRef<Vector2D<int>>& offsets, double* basis)
        {
            if( offsets.size() < 2 )
                std::copy(&(prep(x,y)[0]), &(prep(x,y)[0]) + UnaryBasisSize, basis);
            else
                *basis = 1.0;
        }

        // INTERFACE: return a positive weight for the quadratic part of the specified unary factor
        static double ComputeQuadraticBasis(const PreProcessType& prep, const Vector2D<int>& i, size_t basisIndex)
        {
            return 1.0;
        }

        // INTERFACE: return a positive weight for the quadratic part of the specified pairwise factor
        static double ComputeQuadraticBasis(const PreProcessType& prep, const Vector2D<int>& i, const Vector2D<int>& j, size_t basisIndex)
        {
            return 1.0;
        }

        // INTERFACE: the feature class must have a default constructor
        Feature()
        {
        }

        // INTERFACE: instantiate a new feature with the threshold set from the supplied sample, if applicable
        Feature<TypeTag> WithThresholdFromSample(int x, int y, const PreProcessType& prep,
                const VecCRef<Vector2D<int>>& offsets) const
        {
            return Feature();
        }

        // INTERFACE: decide whether to branch left or right
        bool operator()(int x, int y, const PreProcessType& image,
                        const VecCRef<Vector2D<int>>& offsets) const
        {
            ;
            return true;
        }
    };

    // We only need a feature sampler class for interface compatibility
    template <typename TypeTag = Scene>
    class FeatureSampler
    {
    public:
        FeatureSampler()
        {
        }

        // INTERFACE: the type of features
        typedef Feature<TypeTag> TFeature;

        // INTERFACE: instantiate a new randomly drawn feature
        TFeature operator()(int)
        {
            return TFeature();
        }
    };
}

#endif // H_MULTILABEL_FEATURES_H
