/* This file is part of the "Regression Tree Fields" (RTF) source code distribution,
 * obtained from http://research.microsoft.com/downloads.
 * It is provided to you under the terms of the Microsoft Research License Agreement
 * (MSR-LA). Please see License.txt for details.
 *
 *
 * File: MultiLabel.cpp
 * Implements a multi-label classification example application.
 *
 */

// Define a custom linear operator to be used in this example.
// This should happen before including any headers.
namespace MultiLabel
{
    template<typename TFeature, typename TUnaryGroundLabel> class CompleteInteractionOperator;
}
#define TYPE_COMPLETE_INTERACTION_OPERATOR -1
#define INSTANTIATE_CUSTOM_OPERATOR(typeid) new MultiLabel::CompleteInteractionOperator<TFeature, TUnaryGroundLabel>()

// The use of MPI does not make sense for this example; force-disable it.
#undef USE_MPI

#define LEARNING_USE_PQN

// Regular headers
#include <iostream>

#include <boost/filesystem.hpp>
#include <Eigen/Eigen>

#include "RTF/Loss.h"
#include "RTF/Types.h"
#include "RTF/Basic.h"
#include "RTF/Monitor.h"

#include "Dataset.h"
#include "Features.h"
#include "Operator.h"

// Hyper-parameters to evaluate using crossvalidation
const double C[] = {1e-2, 3e-2, 1e-1, 3e-1, 1e0, 3e0, 1e1, 3e1, 1e2, 3e2, 1e3, 3e3, 1e4, 3e4};
// Number of hyper-parameters
const size_t NumC = sizeof(C)/sizeof(double);
// Perform NumR-fold crossvalidation
const size_t NumR = 10;
// Maximum number of training iterations
size_t MaxNumTrainIt = 1000;
// Stopping condition for training (gradient norm)
double TrainEps = 1e-2;
// Maximum number of inference iterations
const size_t MaxNumInfIt = 500;
// Stopping condition for inference (gradient norm)
const double InfEps = 1e-3;
// Smallest permissible eigenvalue of the positive-definite interaction matrices
const double smallestEigenvalue = 1e-2;

// Determine the validation loss of the model for a given cross-validation run and hyper-parameter C
template<typename DataTag, typename LossTag>
double LossOfCrossValidationRun(const std::string& path, size_t run, double C)
{
    std::cerr << "Crossvalidation run " << run << std::endl;
    const auto NumVariables = MultiLabel::DatasetTraits<DataTag>::NumClasses;

    Basic::RTF<MultiLabel::FeatureSampler<DataTag>,
          MultiLabel::Dataset<DataTag>,
          GradientNormCriterion,
          true,
          false,
          FrobeniusPrior,
          Monitor::NullMonitor> rtf;

    auto op = rtf.AddLinearOperator(TYPE_COMPLETE_INTERACTION_OPERATOR);
    const double params[] = { NumVariables, 1.0/C, smallestEigenvalue, 0 };
    op.SetNonDifferentiableParameters(params);

    MultiLabel::Dataset<DataTag> traindb(path, "train", run, NumR, true);
    MultiLabel::Dataset<DataTag> testdb(path, "train", run, NumR, false);

    std::cerr << "Using " << traindb.GetImageCount() << " examples of the training set for training" << std::endl;
    std::cerr << "Using " << testdb.GetImageCount() << " examples of the training set for testing" << std::endl;

    rtf.template LearnDiscriminative<LossTag, false>(traindb, MaxNumTrainIt, 0, TrainEps, false, MaxNumInfIt, InfEps);

    return rtf.template EvaluateMacroAveraged<Loss::PerPixelError>(testdb, MaxNumInfIt, InfEps);
}

// Via a series of cross-validation runs, determine the best regularization hyper-parameter for the model
template<typename DataTag, typename LossTag>
double FindBestHyperParameter(const std::string& path)
{
    double minLoss = std::numeric_limits<double>::max();
    double minC    = 0;
    for( size_t idxC = 0; idxC < NumC; ++idxC )
    {

        double loss = 0.0;
        for( int idxR = 0; idxR < NumR; ++idxR )
        {
            const auto myloss = LossOfCrossValidationRun<DataTag, LossTag>(path, idxR, C[idxC]);
            //#pragma omp atomic
            loss += myloss;
        }
        loss /= NumR;

        std::cout << "Estimated loss for C = " << C[idxC] << ": " << loss << std::endl;

        if( loss < minLoss )
        {
            minLoss = loss;
            minC    = C[idxC];
        }
    }
    return minC;
}

// Determine the best regularization hyper-parameter via cross-validation and evaluate
// the chosen model on the test data.
template<typename DataTag, typename LossTag>
void Evaluate(const std::string& path)
{
    const auto NumVariables = MultiLabel::DatasetTraits<DataTag>::NumClasses;

    const auto C = FindBestHyperParameter<DataTag, LossTag>(path);
    std::cout << "Picking C = " << C << std::endl;

    Basic::RTF<MultiLabel::FeatureSampler<DataTag>,
          MultiLabel::Dataset<DataTag>,
          GradientNormCriterion,
          true,
          false,
          FrobeniusPrior,
          Monitor::DefaultMonitor> rtf;

    auto op = rtf.AddLinearOperator(TYPE_COMPLETE_INTERACTION_OPERATOR);
    const double params[] = { NumVariables, 1.0/C, smallestEigenvalue, 0 };
    op.SetNonDifferentiableParameters(params);

    MultiLabel::Dataset<DataTag> traindb(path, "train");
    MultiLabel::Dataset<DataTag> testdb(path, "test");

    rtf.template LearnDiscriminative<LossTag, false>(traindb, MaxNumTrainIt, 0, TrainEps, false, MaxNumInfIt, InfEps);
    auto prediction = rtf.Regress(testdb, MaxNumInfIt, InfEps);

    auto meanloss = rtf.template EvaluateMacroAveraged<Loss::PerPixelError>(testdb, prediction);
    std::cout << "Sample mean of the test loss: " << meanloss << std::endl;

    auto meanlosstrain = rtf.template EvaluateMacroAveraged<Loss::PerPixelError>(traindb, MaxNumInfIt, InfEps);
    std::cout << "Sample mean of the train loss: " << meanlosstrain << std::endl;

    double stddev = 0.0;
    size_t numFractional = 0;
    size_t numAll = 0.0;
    for( size_t idx = 0; idx < prediction.size(); ++idx )
    {
        auto pred     = prediction[idx];
        auto thisloss = rtf.template Evaluate<Loss::PerPixelError>(testdb.GetGroundTruthImage(idx), pred);
        stddev += (meanloss - thisloss) * (meanloss - thisloss);

        for( size_t v = 0; v < NumVariables; ++v )
        {
            const auto label = pred(v, 0);
            if( std::fabs(std::fabs(label[0] - label[1]) - 1.0) > 1e-2 )
            {
                numFractional++;
            }
            numAll++;
        }
    }
    stddev = std::sqrt( stddev / (testdb.GetImageCount()-1) );
    std::cout << "Sample standard deviation of the loss: " << stddev << std::endl;

    const auto stderror = stddev / std::sqrt( static_cast<double>(testdb.GetImageCount()) );
    std::cout << "Standard error of the mean: " << stderror << std::endl;

    std::cout << "Percentage of fractional labels: " << ((double)numFractional/numAll)*100.0 << std::endl;
}

int usage()
{
    std::cerr << "Usage: label <loss> <datapath>\n"
              << "\t... where <loss> is one of {logistic,hinge}\n"
              << "\t... where <datapath> points to the dataset directory"
              << std::endl;
    return 1;
}

int main(int argc, char **argv)
{
    if( argc != 3 )
        return usage();

    const std::string loss = argv[1];
    if( loss != "logistic" && loss != "hinge" )
        return usage();

    const std::string path = argv[2];
    if( path.find("scene") == -1 && path.find("yeast") == -1 && path.find("MultiLabel") == -1 )
        return usage();

    try
    {
        if( loss == "logistic" && path.find("scene") != std::string::npos )
            Evaluate<MultiLabel::Scene, Loss::MultiNomialLogistic>(path);
        else if ( loss == "logistic" && path.find("yeast") != std::string::npos )
            Evaluate<MultiLabel::Yeast, Loss::MultiNomialLogistic>(path);
        else if ( loss == "hinge" && path.find("scene") != std::string::npos )
            Evaluate<MultiLabel::Scene, Loss::DiscreteHamming>(path);
        else if ( loss == "hinge" && path.find("yeast") != std::string::npos )
            Evaluate<MultiLabel::Yeast, Loss::DiscreteHamming>(path);
        else if ( loss == "logistic" && path.find("MultiLabel") != std::string::npos )
            Evaluate<MultiLabel::Synth, Loss::MultiNomialLogistic>(path);
        else if ( loss == "hinge" && path.find("MultiLabel") != std::string::npos )
            Evaluate<MultiLabel::Synth, Loss::DiscreteHamming>(path);
        else
            return usage();
    }
    catch ( const std::exception& e )
    {
        std::cerr << "Caught exception: " << e.what() << std::endl;
        return 2;
    }

    return 0;
}
