/* This file is part of the "Regression Tree Fields" (RTF) source code distribution,
 * obtained from http://research.microsoft.com/downloads.
 * It is provided to you under the terms of the Microsoft Research License Agreement
 * (MSR-LA). Please see License.txt for details.
 *
 *
 * File: Operator.h
 * Implements the custom linear operator of the multi-label clasification example.
 *
 */

#ifndef H_COMPLETE_INTERACTION_OPERATOR_H
#define H_COMPLETE_INTERACTION_OPERATOR_H

#include "RTF/Compute.h"
#include "RTF/LinearOperator.h"

namespace MultiLabel
{
    template<typename TFeature, typename TUnaryGroundLabel>
    class CompleteInteractionOperator : public LinearOperator::OperatorBase<TFeature, TUnaryGroundLabel, LinearOperator::DefaultWeights<typename TUnaryGroundLabel::ValueType>>
    {
    public:
        static const size_t VarDim   = TUnaryGroundLabel::Size;
        static const size_t BasisDim = TFeature::UnaryBasisSize;
        typedef typename TUnaryGroundLabel::ValueType TValue;

        typedef Compute::SystemVectorRef<TValue, VarDim>  SystemVectorRef;
        typedef Compute::SystemVectorCRef<TValue, VarDim> SystemVectorCRef;

        typedef Eigen::Matrix<TValue, VarDim, VarDim>   TVarVarMatrix;
        typedef Eigen::Matrix<TValue, VarDim, 1>        TVarVector;
        typedef Eigen::Matrix<TValue, BasisDim, 1>      TBasisVector;
        typedef Eigen::Matrix<TValue, -1, 1>            TVector;

        typedef Compute::Weights<TValue, 2*VarDim, 1>      TPairwiseWeights;
        typedef Compute::Weights<TValue, VarDim, BasisDim> TUnaryWeights;

        typedef Priors::FrobeniusPrior<TValue, 2*VarDim, 1>      TPairwisePrior;
        typedef Priors::FrobeniusPrior<TValue, VarDim, BasisDim> TUnaryPrior;

    private:
        size_t numVariables;
        TValue regularizationC;
        TValue smallestEigenvalue;
        bool unaryOnly;
        VecCRef< Vector2D<int> > offsets;

        mutable std::vector< std::vector<TPairwiseWeights> > pairwiseWeights;
        mutable std::vector< TUnaryWeights > unaryWeights;

    public:
        virtual void AddInImplicitMatrixMultipliedBy(const typename TFeature::PreProcessType& x,
                const SystemVectorRef& Fy, const SystemVectorCRef& y) const
        {
            assert(y.NumPixels() == numVariables);

            for( size_t i = 0; i < numVariables; ++i )
            {

                const auto& var = - unaryWeights[i].Wq;
                Fy[i] += var * y[i];

                for( size_t j = 0; j < numVariables && !unaryOnly; ++j )
                {
                    if( j > i )
                    {
                        const auto& var = - pairwiseWeights[i][j-i-1].Wq.template topLeftCorner < VarDim, VarDim > ();
                        const auto& cov = - pairwiseWeights[i][j-i-1].Wq.template topRightCorner < VarDim, VarDim > ();
                        Fy[i] += var * y[i] + cov * y[j];
                    }
                    else if ( i > j )
                    {
                        const auto& var = - pairwiseWeights[j][i-j-1].Wq.template bottomRightCorner < VarDim, VarDim > ();
                        const auto& cov = - pairwiseWeights[j][i-j-1].Wq.template bottomLeftCorner < VarDim, VarDim > ();
                        Fy[i] += var * y[i] + cov * y[j];
                    }
                }
            }
        }

        virtual void AddInLinearContribution(const typename TFeature::PreProcessType& prep, const SystemVectorRef& l) const
        {
            assert(l.NumPixels() == numVariables);
            const auto cx = l.Width(), cy = l.Height();

            for( int y = 0; y < cy; ++y )
            {
                for( int x = 0; x < cx; ++x )
                {
                    const size_t idx = Compute::VarOffset<1>(cx, cy, x, y);
                    TBasisVector basis;
                    TFeature::ComputeBasis(x, y, prep, offsets, &basis[0]);
                    l(x, y) += unaryWeights[idx].Wl * basis;
                }
            }
        }

        virtual void AccumulateGradient(const typename TFeature::PreProcessType& x,
                                        const SystemVectorCRef& y, const SystemVectorCRef& c, TValue normC) const
        {
            AccumulateQuadraticGradientContribution(x, y, c, normC);
            AccumulateLinearGradientContribution(x, y, c, normC);
        }

        void AccumulateQuadraticGradientContribution(const typename TFeature::PreProcessType& x,
                const SystemVectorCRef& y, const SystemVectorCRef& c, TValue normC) const
        {
            assert(y.NumPixels() == numVariables);

            for( size_t i = 0; i < numVariables; ++i )
            {

                auto& Gvar  = unaryWeights[i].GetGq();
                Gvar += c[i] * (y[i].transpose()/normC);

                for( size_t j = 0; j < numVariables && !unaryOnly; ++j )
                {
                    if( j > i )
                    {
                        auto Gvar = pairwiseWeights[i][j-i-1].GetGq().template topLeftCorner < VarDim, VarDim > ();
                        auto Gcov = pairwiseWeights[i][j-i-1].GetGq().template topRightCorner < VarDim, VarDim > ();

                        Gvar += c[i] * (y[i].transpose()/normC);
                        Gcov += c[i] * (y[j].transpose()/normC);

                    }
                    else if ( i > j )
                    {
                        auto Gvar = pairwiseWeights[j][i-j-1].GetGq().template bottomRightCorner < VarDim, VarDim > ();
                        auto Gcov = pairwiseWeights[j][i-j-1].GetGq().template bottomLeftCorner < VarDim, VarDim > ();

                        Gvar += c[i] * (y[i].transpose()/normC);
                        Gcov += c[i] * (y[j].transpose()/normC);
                    }
                }
            }
        }

        void AccumulateLinearGradientContribution(const typename TFeature::PreProcessType& prep,
                const SystemVectorCRef& y, const SystemVectorCRef& c, TValue normC) const
        {
            assert(y.NumPixels() == numVariables);
            const auto cx = y.Width(), cy = y.Height();

            for( int iy = 0; iy < cy; ++iy )
            {
                for( int ix = 0; ix < cx; ++ix )
                {
                    const size_t idx = Compute::VarOffset<1>(cx, cy, ix, iy);
                    TBasisVector basis;
                    TFeature::ComputeBasis(ix, iy , prep, offsets, &basis[0]);
                    auto& Gl = unaryWeights[idx].GetGl();
                    Gl += c(ix, iy) * (basis.transpose() / normC);
                }
            }
        }

        const TValue* CheckFeasibility(const TValue *ws, bool& feasible) const
        {
            for( int i = 0; i < numVariables; ++i )
            {
                ws = TUnaryWeights::CheckFeasibility(ws, feasible);
                for( int j = i+1; j < numVariables && !unaryOnly; ++j )
                    ws = TPairwiseWeights::CheckFeasibility(ws, feasible);
            }
            return ws;
        }

        TValue* Project(TValue *ws) const
        {
            for( int i = 0; i < numVariables; ++i )
            {
                ws = TUnaryWeights::Project(ws, smallestEigenvalue, 1e10);

                for( int j = i+1; j < numVariables && !unaryOnly; ++j )
                    ws = TPairwiseWeights::Project(ws, smallestEigenvalue, 1e10);
            }
            return ws;
        }

        size_t NumWeights() const
        {
            size_t num = 0;
            for( int i = 0; i < numVariables; ++i )
            {
                num += TUnaryWeights::NumCoefficients;

                for( int j = i+1; j < numVariables && !unaryOnly; ++j )
                    num += TPairwiseWeights::NumCoefficients;
            }
            return num;
        }

        const TValue* SetWeights(const TValue *ws)
        {
            for( int i = 0; i < numVariables; ++i )
            {
                ws = unaryWeights[i].SetWeights(ws);

                for( int j = i+1; j < numVariables && !unaryOnly; ++j )
                    ws = pairwiseWeights[i][j-i-1].SetWeights(ws);
            }
            return ws;
        }

        TValue* GetWeights(TValue *ws) const
        {
            for( int i = 0; i < numVariables; ++i )
            {
                ws = unaryWeights[i].GetWeights(ws);

                for( int j = i+1; j < numVariables && !unaryOnly; ++j )
                    ws = pairwiseWeights[i][j-i-1].GetWeights(ws);
            }
            return ws;
        }

        TValue* GetGradientAddPrior(TValue *gs, TValue& objective) const
        {
            for( int i = 0; i < numVariables; ++i )
            {
                objective += TUnaryPrior::ComputeObjectiveAddGradient(regularizationC, regularizationC, unaryWeights[i]);
                gs = unaryWeights[i].GetGradient(gs);

                for( int j = i+1; j < numVariables && !unaryOnly; ++j )
                {
                    objective += TPairwisePrior::ComputeObjectiveAddGradient(regularizationC, regularizationC, pairwiseWeights[i][j-i-1]);
                    gs = pairwiseWeights[i][j-i-1].GetGradient(gs);
                }
            }
            return gs;
        }

        void ClearGradient()
        {
            for( int i = 0; i < numVariables; ++i )
            {
                unaryWeights[i].ClearGradient();

                for( int j = i+1; j < numVariables && !unaryOnly; ++j )
                    pairwiseWeights[i][j-i-1].ClearGradient();
            }
        }

        const TValue* SetNonDifferentiableParameters(const TValue* flatWeights)
        {
            // Set non-differentiable parameters of operator
            numVariables = static_cast<size_t>(*flatWeights++);
            regularizationC = *flatWeights++;
            smallestEigenvalue = *flatWeights++;
            unaryOnly = (*flatWeights++ != 0.0);

            // Set up weights
            pairwiseWeights.resize(numVariables);
            for( int i = 0; i < numVariables; ++i )
            {
                unaryWeights.push_back(TUnaryWeights());
                for( int j = i+1; j < numVariables && !unaryOnly; ++j )
                    pairwiseWeights[i].push_back(TPairwiseWeights());
            }
            return flatWeights;
        }

        size_t NumNonDifferentiableParameters() const
        {
            return 4;
        }

        void ResetWeights()
        {
            for( int i = 0; i < numVariables; ++i )
            {
                unaryWeights[i].Reset(smallestEigenvalue);

                for( int j = i+1; j < numVariables && !unaryOnly; ++j )
                    pairwiseWeights[i][j-i-1].Reset(smallestEigenvalue);
            }
        }

        void Print() const
        {
        }

        int Type() const
        {
            return TYPE_COMPLETE_INTERACTION_OPERATOR;
        }

        std::istream& Deserialize(std::istream& in)
        {
            return in;
        }

        std::ostream& Serialize(std::ostream& out) const
        {
            return out;
        }

#ifdef _OPENMP
        // We don't need a lock
        void InitializeLocks()
        {
        }

        void DestroyLocks()
        {
        }
#endif
    };
}

#endif // H_COMPLETE_INTERACTION_OPERATOR_H
